#ifndef bqtQBlibStreamHH
#define bqtQBlibStreamHH

#include <qblib/types.hh>

#include <ostream>
#include <istream>
#include <cstdio>

class FBufType : public std::streambuf
{
public:
    typedef std::streambuf::int_type int_type;

    FBufType (FILE* f) : std::streambuf(), fp(f)
    {
    }

    ~FBufType()
    {
    }

    int_type underflow() { return underflow_common (false); }
    int_type uflow()     { return underflow_common (true); }

    int_type overflow (int_type c)
    {
        return (c != traits_type::eof()) ? fputc(c, fp) : flush();
    }

    int_type pbackfail (int_type c)
    {
        return (c != EOF) ? ungetc(c, fp) : traits_type::not_eof(c);
    }

    std::streamsize xsputn(const char* src, std::streamsize n)
    {
        return fwrite(src, 1, n, fp);
    }

    std::streamsize xsgetn(char *dest, std::streamsize n)
    {
        return fread(dest, 1, n, fp);
    }

    std::streampos seekoff (std::streamoff ofs, std::ios::seekdir dir,
                std::ios::openmode = std::ios::in | std::ios::out)
    {
        switch(dir)
        {
            default:
            case std::ios::beg:
                return fseek(fp, ofs, SEEK_SET);
            case std::ios::cur:
                return fseek(fp, ofs, SEEK_CUR);
            case std::ios::end:
                return fseek(fp, ofs, SEEK_END);
        }
    }

    std::streampos seekpos (std::streampos pos,
                std::ios::openmode = std::ios::in | std::ios::out)
    {
        return seekoff(pos, std::ios::beg);
    }
    

protected:
    FILE* fp;

private:
    int flush() { return fflush(fp); }
    int_type underflow_common (bool bump)
    {
        int c = fgetc(fp);
        if(!bump && c != EOF)  ungetc(c, fp);
        return c;
    }
};

class PrintStreamBase: public std::ostream
{
private:
    FBufType Buf;
public:
    PrintStreamBase(FILE*f) : std::ostream(), Buf(f) { }
    
    virtual ~PrintStreamBase() {}
    
    virtual void DoSPC(int count) {};
    virtual void DoTAB(int count) {};
    virtual void DoNextTAB() {};
    virtual void DoCRLF() {};

    FBufType* rdbuf() const { return const_cast<FBufType*> (&Buf); }
};

class InputStreamBase: public std::istream
{
private:
    FBufType Buf;
public:
    InputStreamBase(FILE*f) : std::istream(), Buf(f) { }
    
    virtual ~InputStreamBase() {}
    
    virtual void ReadLine(BAS_STR& target) {};
    virtual void InputTo(BAS_INT& target) {};
    virtual void InputTo(BAS_DBL& target) {};
    virtual void InputTo(BAS_STR& target) {};
    virtual void InputTo(BAS_SGN& target) {};
    virtual void InputTo(BAS_LNG& target) {};
    virtual bool Satisfied() {};

    virtual BAS_STR Input_N(unsigned n) {};
    virtual BAS_STR Inkey_S() {};
    
    FBufType* rdbuf() const { return const_cast<FBufType*> (&Buf); }
    
    static inline const char* GetRedoMsg() { return "\n?Redo from start;\n"; }
};

extern PrintStreamBase& GET_STREAM_OUTPUT(int filenumber);
extern PrintStreamBase& GET_PRINTER();
extern PrintStreamBase& GET_STDOUT();

extern InputStreamBase& GET_STDIN();
extern InputStreamBase& GET_STREAM_INPUT(int filenumber);

class UsingFormatter
{
public:
    template<typename T>
    void DoParam(PrintStreamBase& out, const T& value);
    /* Note: DoParam() can execute B_ERROR(13) (Type mismatch) */
};


static inline const BAS_STR B_FuncINPUT_S(int n)
    { return GET_STDIN().Input_N(n); }
static inline const BAS_STR B_FuncINPUT_S(int n, int filenumber)
    { return GET_STREAM_INPUT(filenumber).Input_N(n); }
static inline const BAS_STR B_FuncINKEY_S()
    { return GET_STDIN().Inkey_S(); }

#endif
